#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community",
}

DOCUMENTATION = """
---
module: cp_mgmt_add_data_center_query
short_description: Adds data-center-query objects on Checkpoint over Web Services API
description:
  - Adds data-center-query objects on Checkpoint devices.
  - All operations are performed over Web Services API.
  - Available from R81 management version.
version_added: "6.9.0"
author: "Dor Berenstein (@chkp-dorbe)"
options:
  name:
    description:
      - Object name.
    type: str
    required: True
  query_rules:
    description:
      - Data Center Query Rules.<br>There is an 'AND' operation between multiple Query Rules.
    type: list
    elements: dict
    suboptions:
      key_type:
        description:
          - The type of the "key" parameter.<br>Use "predefined" for these keys, type-in-data-center, name-in-data-center, and ip-address.<br>Use
            "tag" to query the Data Center tag's property.
        type: str
        choices: ['predefined', 'tag']
      key:
        description:
          - Defines in which Data Center property to query.<br>For key-type "predefined", use these keys, type-in-data-center,
            name-in-data-center, and ip-address.<br>For key-type "tag", use the Data Center tag key to query.<br>Keys are case-insensitive.
        type: str
      values:
        description:
          - The value(s) of the Data Center property to match the Query Rule.<br>Values are case-insensitive.<br>There is an 'OR' operation
            between multiple values.<br>For key-type "predefined" and key 'ip-address', the values must be an IPv4 or IPv6 address.<br>For key-type "tag", the
            values must be the Data Center tag values.
        type: list
        elements: str
  data_centers:
    description:
      - Collection of Data Center servers identified by the name or UID. use "All" to select all data centers.
    type: list
    elements: str
  color:
    description:
      - Color of the object. Should be one of existing colors.
    type: str
    choices: ['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green', 'khaki', 'orchid', 'dark orange', 'dark sea green',
             'pink', 'turquoise', 'dark blue', 'firebrick', 'brown', 'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon',
             'coral', 'sea green', 'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna', 'yellow']
  comments:
    description:
      - Comments string.
    type: str
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  tags:
    description:
      - Collection of tag identifiers.
    type: list
    elements: str
  ignore_warnings:
    description:
      - Apply changes ignoring warnings.
    type: bool
  ignore_errors:
    description:
      - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_commands
"""

EXAMPLES = """
- name: add-data-center-query
  cp_mgmt_add_data_center_query:
    name: data-center-query1
"""

RETURN = """
cp_mgmt_add_data_center_query:
  description: The checkpoint add-data-center-query output.
  returned: always.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import (
    checkpoint_argument_spec_for_commands,
    api_command,
)


def main():
    argument_spec = dict(
        name=dict(type="str", required=True),
        query_rules=dict(
            type="list",
            elements="dict",
            options=dict(
                key_type=dict(type="str", choices=["predefined", "tag"]),
                key=dict(type="str", no_log=False),
                values=dict(type="list", elements="str"),
            ),
        ),
        data_centers=dict(type="list", elements="str"),
        color=dict(
            type="str",
            choices=[
                "aquamarine",
                "black",
                "blue",
                "crete blue",
                "burlywood",
                "cyan",
                "dark green",
                "khaki",
                "orchid",
                "dark orange",
                "dark sea green",
                "pink",
                "turquoise",
                "dark blue",
                "firebrick",
                "brown",
                "forest green",
                "gold",
                "dark gold",
                "gray",
                "dark gray",
                "light green",
                "lemon chiffon",
                "coral",
                "sea green",
                "sky blue",
                "magenta",
                "purple",
                "slate blue",
                "violet red",
                "navy blue",
                "olive",
                "orange",
                "red",
                "sienna",
                "yellow",
            ],
        ),
        comments=dict(type="str"),
        details_level=dict(type="str", choices=["uid", "standard", "full"]),
        tags=dict(type="list", elements="str"),
        ignore_warnings=dict(type="bool"),
        ignore_errors=dict(type="bool"),
    )
    argument_spec.update(checkpoint_argument_spec_for_commands)

    module = AnsibleModule(argument_spec=argument_spec)

    command = "add-data-center-query"

    result = api_command(module, command)
    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: split_pem
short_description: Split PEM file contents into multiple objects
version_added: 2.10.0
author:
  - Felix Fontein (@felixfontein)
description:
  - Split PEM file contents into multiple PEM objects. Comments or invalid parts are ignored.
options:
  _input:
    description:
      - The PEM contents to split.
    type: string
    required: true
"""

EXAMPLES = r"""
---
- name: Print all CA certificates
  ansible.builtin.debug:
    msg: '{{ item }}'
  loop: >-
    {{ lookup('ansible.builtin.file', '/path/to/ca-bundle.pem') | community.crypto.split_pem }}
"""

RETURN = r"""
_value:
  description:
    - A list of PEM file contents.
  type: list
  elements: string
"""

from collections.abc import Callable

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.crypto.plugins.module_utils._crypto.pem import (
    split_pem_list,
)


def split_pem_filter(data: str | bytes) -> list[str]:
    """Split PEM file."""
    if not isinstance(data, (str, bytes)):
        raise AnsibleFilterError(
            f"The community.crypto.split_pem input must be a text type, not {type(data)}"
        )

    return split_pem_list(to_text(data))


class FilterModule:
    """Ansible jinja2 filters"""

    def filters(self) -> dict[str, Callable]:
        return {
            "split_pem": split_pem_filter,
        }



# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: doas
short_description: Do As user
description:
  - This become plugins allows your remote/login user to execute commands as another user using the C(doas) utility.
author: Ansible Core Team
options:
  become_user:
    description: User you 'become' to execute the task.
    type: string
    ini:
      - section: privilege_escalation
        key: become_user
      - section: doas_become_plugin
        key: user
    vars:
      - name: ansible_become_user
      - name: ansible_doas_user
    env:
      - name: ANSIBLE_BECOME_USER
      - name: ANSIBLE_DOAS_USER
  become_exe:
    description: C(doas) executable.
    type: string
    default: doas
    ini:
      - section: privilege_escalation
        key: become_exe
      - section: doas_become_plugin
        key: executable
    vars:
      - name: ansible_become_exe
      - name: ansible_doas_exe
    env:
      - name: ANSIBLE_BECOME_EXE
      - name: ANSIBLE_DOAS_EXE
  become_flags:
    description: Options to pass to C(doas).
    type: string
    default: ''
    ini:
      - section: privilege_escalation
        key: become_flags
      - section: doas_become_plugin
        key: flags
    vars:
      - name: ansible_become_flags
      - name: ansible_doas_flags
    env:
      - name: ANSIBLE_BECOME_FLAGS
      - name: ANSIBLE_DOAS_FLAGS
  become_pass:
    description: Password for C(doas) prompt.
    type: string
    required: false
    vars:
      - name: ansible_become_password
      - name: ansible_become_pass
      - name: ansible_doas_pass
    env:
      - name: ANSIBLE_BECOME_PASS
      - name: ANSIBLE_DOAS_PASS
    ini:
      - section: doas_become_plugin
        key: password
  prompt_l10n:
    description:
      - List of localized strings to match for prompt detection.
      - If empty the plugin uses the built-in one.
    type: list
    elements: string
    default: []
    ini:
      - section: doas_become_plugin
        key: localized_prompts
    vars:
      - name: ansible_doas_prompt_l10n
    env:
      - name: ANSIBLE_DOAS_PROMPT_L10N
  allow_pipelining:
    description:
      - When set to V(true), do allow pipelining with ansible-core 2.19+.
      - This should only be used when doas is configured to not ask for a password (C(nopass)).
    type: boolean
    default: false
    version_added: 12.4.0
    ini:
      - section: doas_become_plugin
        key: allow_pipelining
    vars:
      - name: ansible_doas_allow_pipelining
    env:
      - name: ANSIBLE_DOAS_ALLOW_PIPELINING
notes:
  - This become plugin does not work when connection pipelining is enabled
    and doas requests a password.
    With ansible-core 2.19+, using this plugin automatically disables pipelining,
    unless O(allow_pipelining=true) is explicitly set by the user.
    On ansible-core 2.18 and before, pipelining must explicitly be disabled by the user.
"""

import re

from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.become import BecomeBase


class BecomeModule(BecomeBase):
    name = "community.general.doas"

    # messages for detecting prompted password issues
    fail = ("Permission denied",)
    missing = ("Authorization required",)

    # See https://github.com/ansible-collections/community.general/issues/9977,
    # https://github.com/ansible/ansible/pull/78111,
    # https://github.com/ansible-collections/community.general/issues/11411
    @property
    def pipelining(self) -> bool:  # type: ignore[override]
        return self.get_option("allow_pipelining")

    def check_password_prompt(self, b_output):
        """checks if the expected password prompt exists in b_output"""

        # FIXME: more accurate would be: 'doas (%s@' % remote_user
        # however become plugins don't have that information currently
        b_prompts = [to_bytes(p) for p in self.get_option("prompt_l10n")] or [rb"doas \(", rb"Password:"]
        b_prompt = b"|".join(b_prompts)

        return bool(re.match(b_prompt, b_output))

    def build_become_command(self, cmd, shell):
        super().build_become_command(cmd, shell)

        if not cmd:
            return cmd

        self.prompt = True

        become_exe = self.get_option("become_exe")

        flags = self.get_option("become_flags")
        if not self.get_option("become_pass") and "-n" not in flags:
            flags += " -n"

        become_user = self.get_option("become_user")
        user = f"-u {become_user}" if become_user else ""

        success_cmd = self._build_success_command(cmd, shell, noexe=True)
        executable = getattr(shell, "executable", shell.SHELL_FAMILY)

        return f"{become_exe} {flags} {user} {executable} -c {success_cmd}"



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_customer_gateway_info
version_added: 1.0.0
short_description: Gather information about customer gateways in AWS
description:
  - Gather information about customer gateways in AWS.
author:
  - Madhura Naniwadekar (@Madhura-CSI)
options:
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value.
        See U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeCustomerGateways.html) for possible filters.
    type: dict
    default: {}
  customer_gateway_ids:
    description:
      - Get details of a specific customer gateways using customer gateway ID/IDs. This value should be provided as a list.
    type: list
    elements: str
    default: []
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# # Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Gather information about all customer gateways
  community.aws.ec2_customer_gateway_info:

- name: Gather information about a filtered list of customer gateways, based on tags
  community.aws.ec2_customer_gateway_info:
    region: ap-southeast-2
    filters:
      "tag:Name": test-customer-gateway
      "tag:AltName": test-customer-gateway-alt
  register: cust_gw_info

- name: Gather information about a specific customer gateway by specifying customer gateway ID
  community.aws.ec2_customer_gateway_info:
    region: ap-southeast-2
    customer_gateway_ids:
      - 'cgw-48841a09'
      - 'cgw-fec021ce'
  register: cust_gw_info
"""

RETURN = r"""
customer_gateways:
    description: List of one or more customer gateways.
    returned: always
    type: list
    sample: [
            {
                "bgp_asn": "65000",
                "customer_gateway_id": "cgw-fec844ce",
                "customer_gateway_name": "test-customer-gw",
                "ip_address": "110.112.113.120",
                "state": "available",
                "tags": [
                    {
                        "key": "Name",
                        "value": "test-customer-gw"
                    }
                ],
                "type": "ipsec.1"
            }
        ]
"""

import json

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def date_handler(obj):
    return obj.isoformat() if hasattr(obj, "isoformat") else obj


def list_customer_gateways(connection, module):
    params = dict()

    params["Filters"] = ansible_dict_to_boto3_filter_list(module.params.get("filters"))
    params["CustomerGatewayIds"] = module.params.get("customer_gateway_ids")

    try:
        result = json.loads(json.dumps(connection.describe_customer_gateways(**params), default=date_handler))
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg="Could not describe customer gateways")
    snaked_customer_gateways = [camel_dict_to_snake_dict(gateway) for gateway in result["CustomerGateways"]]
    if snaked_customer_gateways:
        for customer_gateway in snaked_customer_gateways:
            customer_gateway["tags"] = boto3_tag_list_to_ansible_dict(customer_gateway.get("tags", []))
            customer_gateway_name = customer_gateway["tags"].get("Name")
            if customer_gateway_name:
                customer_gateway["customer_gateway_name"] = customer_gateway_name
    module.exit_json(changed=False, customer_gateways=snaked_customer_gateways)


def main():
    argument_spec = dict(
        customer_gateway_ids=dict(default=[], type="list", elements="str"), filters=dict(default={}, type="dict")
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["customer_gateway_ids", "filters"],
        ],
        supports_check_mode=True,
    )

    connection = module.client("ec2")

    list_customer_gateways(connection, module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the nxos_ntp_global module
"""


class Ntp_globalArgs(object):  # pylint: disable=R0903
    """The arg spec for the nxos_ntp_global module"""

    argument_spec = {
        "running_config": {"type": "str"},
        "config": {
            "type": "dict",
            "options": {
                "access_group": {
                    "type": "dict",
                    "options": {
                        "match_all": {"type": "bool"},
                        "peer": {
                            "type": "list",
                            "elements": "dict",
                            "options": {"access_list": {"type": "str"}},
                        },
                        "query_only": {
                            "type": "list",
                            "elements": "dict",
                            "options": {"access_list": {"type": "str"}},
                        },
                        "serve": {
                            "type": "list",
                            "elements": "dict",
                            "options": {"access_list": {"type": "str"}},
                        },
                        "serve_only": {
                            "type": "list",
                            "elements": "dict",
                            "options": {"access_list": {"type": "str"}},
                        },
                    },
                },
                "allow": {
                    "type": "dict",
                    "options": {
                        "control": {
                            "type": "dict",
                            "options": {"rate_limit": {"type": "int"}},
                        },
                        "private": {"type": "bool"},
                    },
                },
                "authenticate": {"type": "bool"},
                "authentication_keys": {
                    "type": "list",
                    "elements": "dict",
                    "no_log": False,
                    "options": {
                        "id": {"type": "int"},
                        "key": {"type": "str", "no_log": True},
                        "encryption": {"type": "int"},
                    },
                },
                "logging": {"type": "bool"},
                "master": {
                    "type": "dict",
                    "options": {"stratum": {"type": "int"}},
                },
                "passive": {"type": "bool"},
                "peers": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "peer": {"type": "str"},
                        "key_id": {"type": "int"},
                        "maxpoll": {"type": "int"},
                        "minpoll": {"type": "int"},
                        "prefer": {"type": "bool"},
                        "vrf": {"type": "str", "aliases": ["use_vrf"]},
                    },
                },
                "servers": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "server": {"type": "str"},
                        "key_id": {"type": "int"},
                        "maxpoll": {"type": "int"},
                        "minpoll": {"type": "int"},
                        "prefer": {"type": "bool"},
                        "vrf": {"type": "str", "aliases": ["use_vrf"]},
                    },
                },
                "source": {"type": "str"},
                "source_interface": {"type": "str"},
                "trusted_keys": {
                    "type": "list",
                    "elements": "dict",
                    "no_log": False,
                    "options": {"key_id": {"type": "int"}},
                },
            },
        },
        "state": {
            "type": "str",
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "parsed",
                "gathered",
                "rendered",
            ],
            "default": "merged",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_vlan_policy_info
short_description: Gather information about VLAN Policies and their VLANs in Cisco Intersight
description:
  - Retrieve comprehensive information about VLAN Policies and their associated VLANs from L(Cisco Intersight,https://intersight.com).
  - Query policies by organization, policy name, or filter VLANs by name patterns.
  - Returns structured data combining policy metadata with detailed VLAN configurations.
  - Supports filtering by organization, policy name, and VLAN name.
  - If no filters are provided, all VLAN Policies and their VLANs will be returned.
  - Returns structured data with both policy information and associated VLAN details.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization to filter VLAN Policies by.
      - Use 'default' for the default organization.
      - When specified, only policies from this organization will be returned.
    type: str
  name:
    description:
      - The exact name of the VLAN Policy to retrieve information from.
      - When specified, only the matching policy and its VLANs will be returned.
      - Can be combined with vlan_name filter to get specific VLANs from a specific policy.
    type: str
  vlan_name:
    description:
      - Filter VLANs by name within the policies.
      - Can be used with or without policy name filtering.
      - Supports exact matching (e.g., "prod_100" will match only "prod_100").
      - When combined with policy name, filters VLANs within that specific policy.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
# Basic Usage Examples
- name: Fetch all VLAN Policies and their VLANs from all organizations
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
  register: all_vlan_policies

# Organization-specific Examples
- name: Fetch all VLAN Policies from the default organization
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
  register: default_org_policies

- name: Fetch all VLAN Policies from a custom organization
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "Engineering"
  register: engineering_policies

# Specific Policy Examples
- name: Fetch a specific VLAN Policy by name with all its VLANs
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "datacenter-vlan-policy"
  register: specific_policy

- name: Display policy details
  debug:
    msg: "Policy {{ specific_policy.api_response.policy.Name }} has {{ specific_policy.api_response.vlans | length }} VLANs"

# VLAN Filtering Examples
- name: Find all policies containing a specific VLAN
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    vlan_name: "prod_100"
  register: policies_with_prod_100

- name: Find specific VLAN in a specific policy
  cisco.intersight.intersight_vlan_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "datacenter-vlan-policy"
    vlan_name: "mgmt_300"
  register: specific_vlan_in_policy
'''

RETURN = r'''
api_response:
  description:
    - The API response containing policy and VLAN information.
    - Returns a dictionary when querying a single policy or no policies found.
    - Returns a list when multiple policies are found.
  returned: always
  type: dict
  contains:
    vlan_policy:
      description: VLAN policy information
      type: dict
      returned: always
    vlans:
      description: List of VLANs associated with the policy
      type: list
      returned: always
  sample:
    # Single policy response (when name parameter is used or only one policy found)
    vlan_policy:
      Name: "datacenter-vlan-policy"
      ObjectType: "fabric.EthNetworkPolicy"
      Moid: "12345678901234567890abcd"
      Description: "VLAN policy for datacenter infrastructure"
      Organization:
        Name: "default"
        ObjectType: "organization.Organization"
      Tags:
        - Key: "Site"
          Value: "DataCenter-A"
        - Key: "Environment"
          Value: "Production"
    vlans:
      - Name: "prod_100"
        ObjectType: "fabric.Vlan"
        Moid: "vlan12345678901234567890"
        VlanId: 100
        AutoAllowOnUplinks: true
        IsNative: false
        SharingType: "None"
        MulticastPolicy:
          Name: "default-multicast-policy"
          ObjectType: "fabric.MulticastPolicy"
      - Name: "dmz_primary_50"
        ObjectType: "fabric.Vlan"
        Moid: "vlan09876543210987654321"
        VlanId: 50
        AutoAllowOnUplinks: true
        IsNative: false
        SharingType: "Primary"
        PrimaryVlanId: 0
      - Name: "dmz_isolated_51"
        ObjectType: "fabric.Vlan"
        Moid: "vlanabcdef1234567890123456"
        VlanId: 51
        AutoAllowOnUplinks: true
        IsNative: false
        SharingType: "Isolated"
        PrimaryVlanId: 50
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def get_vlans_for_policy(intersight, policy_moid, vlan_name_filter=None):
    """
    Get all VLANs associated with a specific policy.

    Args:
        intersight: IntersightModule instance
        policy_moid: MOID of the policy to get VLANs for
        vlan_name_filter: Optional filter for VLAN names

    Returns:
        List of VLAN objects associated with the policy
    """
    # Build filter for VLANs associated with this policy
    filter_conditions = [f"EthNetworkPolicy.Moid eq '{policy_moid}'"]

    # Add VLAN name filter if provided (exact match only)
    if vlan_name_filter:
        filter_conditions.append(f"Name eq '{vlan_name_filter}'")

    query_params = {
        '$filter': ' and '.join(filter_conditions)
    }

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get VLANs for this policy
    intersight.get_resource(
        resource_path='/fabric/Vlans',
        query_params=query_params,
        return_list=True
    )

    # Capture the response immediately before it gets overwritten
    vlans_response = intersight.result.get('api_response', [])

    # Ensure we return a list
    if isinstance(vlans_response, list):
        return vlans_response
    elif isinstance(vlans_response, dict):
        return [vlans_response]
    else:
        return []


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str'),
        vlan_name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)

    # Resource path used to fetch policy info
    resource_path = '/fabric/EthNetworkPolicies'

    # Get query parameters for policies
    query_params = intersight.set_query_params()

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get VLAN policies
    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    policies = intersight.result.get('api_response', [])
    vlan_name_filter = module.params.get('vlan_name')

    # Build structured response
    structured_results = []

    # Ensure policies is always a list for iteration, even if a single dict is returned
    if isinstance(policies, dict):
        policies = [policies]
    elif not isinstance(policies, list):
        policies = []

    for policy in policies:
        policy_moid = policy.get('Moid')
        if not policy_moid:
            continue

        # Get VLANs for this policy
        vlans = get_vlans_for_policy(intersight, policy_moid, vlan_name_filter)

        # Create structured response for this policy
        policy_result = {
            'vlan_policy': policy,
            'vlans': vlans
        }

        structured_results.append(policy_result)

    # Create final response structure
    final_api_response = None

    # Set final response based on number of policies found
    if len(structured_results) == 1:
        # Single policy - return as dict
        final_api_response = structured_results[0]
    elif len(structured_results) > 1:
        # Multiple policies - return as list
        final_api_response = structured_results
    else:
        # No policies found - return empty structure
        final_api_response = {
            'vlan_policy': {},
            'vlans': []
        }

    # Use intersight.result and update api_response directly
    intersight.result['api_response'] = final_api_response
    final_result = intersight.result

    module.exit_json(**final_result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2014, Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: nexmo
short_description: Send a SMS using nexmo
description:
  - Send a SMS message using nexmo.
author: "Matt Martz (@sivel)"
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  api_key:
    type: str
    description:
      - Nexmo API Key.
    required: true
  api_secret:
    type: str
    description:
      - Nexmo API Secret.
    required: true
  src:
    type: int
    description:
      - Nexmo Number to send from.
    required: true
  dest:
    type: list
    elements: int
    description:
      - Phone number(s) to send SMS message to.
    required: true
  msg:
    type: str
    description:
      - Message text to send. Messages longer than 160 characters are split into multiple messages.
    required: true
  validate_certs:
    description:
      - If V(false), SSL certificates are not validated. This should only be used on personally controlled sites using self-signed
        certificates.
    type: bool
    default: true
extends_documentation_fragment:
  - ansible.builtin.url
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Send notification message via Nexmo
  community.general.nexmo:
    api_key: 640c8a53
    api_secret: 0ce239a6
    src: 12345678901
    dest:
      - 10987654321
      - 16789012345
    msg: '{{ inventory_hostname }} completed'
  delegate_to: localhost
"""
import json
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url, url_argument_spec

NEXMO_API = "https://rest.nexmo.com/sms/json"


def send_msg(module):
    failed = list()
    responses = dict()
    msg = {
        "api_key": module.params.get("api_key"),
        "api_secret": module.params.get("api_secret"),
        "from": module.params.get("src"),
        "text": module.params.get("msg"),
    }
    for number in module.params.get("dest"):
        msg["to"] = number
        url = f"{NEXMO_API}?{urlencode(msg)}"

        headers = dict(Accept="application/json")
        response, info = fetch_url(module, url, headers=headers)
        if info["status"] != 200:
            failed.append(number)
            responses[number] = dict(failed=True)

        try:
            responses[number] = json.load(response)
        except Exception:
            failed.append(number)
            responses[number] = dict(failed=True)
        else:
            for message in responses[number]["messages"]:
                if int(message["status"]) != 0:
                    failed.append(number)
                    responses[number] = dict(failed=True, **responses[number])

        if failed:
            msg = "One or messages failed to send"
        else:
            msg = ""

        module.exit_json(failed=bool(failed), msg=msg, changed=False, responses=responses)


def main():
    argument_spec = url_argument_spec()
    argument_spec.update(
        dict(
            api_key=dict(required=True, no_log=True),
            api_secret=dict(required=True, no_log=True),
            src=dict(required=True, type="int"),
            dest=dict(required=True, type="list", elements="int"),
            msg=dict(required=True),
        ),
    )

    module = AnsibleModule(argument_spec=argument_spec)

    send_msg(module)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
"""
Shared utilities for filter plugins that use AnsibleArgSpecValidator.

In Ansible 2.19+, filter arguments may be wrapped in lazy containers from
ansible-core (lib/ansible/_internal/_templating/_lazy_containers.py). Those
containers subclass dict/list/tuple and resolve values on iteration (e.g.
``.items()``, ``__iter__``). They implement ``__deepcopy__`` but it deepcopies
``_templar`` and ``_lazy_options``, which can fail; ansible-core's
ArgumentSpecValidator (lib/ansible/module_utils/common/arg_spec.py) uses
deepcopy in ValidationResult, so passing lazy containers can raise. Converting
filter args to native types before AnsibleArgSpecValidator avoids that. This
implementation aligns with the lazy containers: we recurse over dict/list/tuple
(so lazy subclasses are iterated and thus resolved) and build plain native
containers. ansible-core does not currently expose a public API for "lazy to
native" conversion; the only mechanism is the private ``_non_lazy_copy()`` on
lazy container instances (ansible._internal._templating._lazy_containers). We
avoid importing from ``_internal`` so collections remain compatible with
future core changes; if core adds a public API later, this helper can be
updated to use it.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


def convert_to_native(value):
    """Convert Ansible lazy containers and wrapped types to native Python types.

    Use this for filter plugin argument dicts before passing them to
    AnsibleArgSpecValidator. Lazy containers (see ansible-core
    ``_lazy_containers.py``) subclass dict/list/tuple and resolve on iteration;
    this function iterates and recurses so resolved values become plain dicts,
    lists, and scalars, avoiding deepcopy failures in ArgumentSpecValidator.

    Recommendation: keep this implementation and avoid importing from
    ansible._internal (e.g. _non_lazy_copy()). If ansible-core adds a public
    "lazy to native" API later, this helper can be updated to use it.
    """
    import json

    if value is None:
        return None
    if isinstance(value, (bool, int, float, str)):
        return value
    if isinstance(value, (list, tuple)):
        return [convert_to_native(item) for item in value]
    if isinstance(value, dict):
        return {convert_to_native(k): convert_to_native(v) for k, v in value.items()}
    # For any other type, try to convert via JSON round-trip to get native types
    try:
        return json.loads(json.dumps(value))
    except (TypeError, ValueError):
        return value



#!/usr/bin/python
#
# Copyright (C) 2019 Huawei
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

###############################################################################
# Documentation
###############################################################################

DOCUMENTATION = r"""
module: hwc_ecs_instance
description:
  - Instance management.
short_description: Creates a resource of Ecs/Instance in Huawei Cloud
version_added: '0.2.0'
author: Huawei Inc. (@huaweicloud)
requirements:
  - keystoneauth1 >= 3.6.0
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether the given object should exist in Huawei Cloud.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  timeouts:
    description:
      - The timeouts for each operations.
    type: dict
    default: {}
    suboptions:
      create:
        description:
          - The timeouts for create operation.
        type: str
        default: '30m'
      update:
        description:
          - The timeouts for update operation.
        type: str
        default: '30m'
      delete:
        description:
          - The timeouts for delete operation.
        type: str
        default: '30m'
  availability_zone:
    description:
      - Specifies the name of the AZ where the ECS is located.
    type: str
    required: true
  flavor_name:
    description:
      - Specifies the name of the system flavor.
    type: str
    required: true
  image_id:
    description:
      - Specifies the ID of the system image.
    type: str
    required: true
  name:
    description:
      - Specifies the ECS name. Value requirements consists of 1 to 64 characters, including letters, digits, underscores
        (V(_)), hyphens (V(-)), periods (V(.)).
    type: str
    required: true
  nics:
    description:
      - Specifies the NIC information of the ECS. Constraints the network of the NIC must belong to the VPC specified by vpc_id.
        A maximum of 12 NICs can be attached to an ECS.
    type: list
    elements: dict
    required: true
    suboptions:
      ip_address:
        description:
          - Specifies the IP address of the NIC. The value is an IPv4 address. Its value must be an unused IP address in the
            network segment of the subnet.
        type: str
        required: true
      subnet_id:
        description:
          - Specifies the ID of subnet.
        type: str
        required: true
  root_volume:
    description:
      - Specifies the configuration of the ECS's system disks.
    type: dict
    required: true
    suboptions:
      volume_type:
        description:
          - Specifies the ECS system disk type.
          - SATA is common I/O disk type.
          - SAS is high I/O disk type.
          - SSD is ultra-high I/O disk type.
          - Co-p1 is high I/O (performance-optimized I) disk type.
          - Uh-l1 is ultra-high I/O (latency-optimized) disk type.
          - NOTE is For HANA, HL1, and HL2 ECSs, use co-p1 and uh-l1 disks. For other ECSs, do not use co-p1 or uh-l1 disks.
        type: str
        required: true
      size:
        description:
          - Specifies the system disk size, in GB. The value range is 1 to 1024. The system disk size must be greater than
            or equal to the minimum system disk size supported by the image (min_disk attribute of the image). If this parameter
            is not specified or is set to 0, the default system disk size is the minimum value of the system disk in the image
            (min_disk attribute of the image).
        type: int
      snapshot_id:
        description:
          - Specifies the snapshot ID or ID of the original data disk contained in the full-ECS image.
        type: str
  vpc_id:
    description:
      - Specifies the ID of the VPC to which the ECS belongs.
    type: str
    required: true
  admin_pass:
    description:
      - Specifies the initial login password of the administrator account for logging in to an ECS using password authentication.
        The Linux administrator is root, and the Windows administrator is Administrator. Password complexity requirements,
        consists of 8 to 26 characters. The password must contain at least three of the following character types 'uppercase
        letters, lowercase letters, digits, and special characters (V(!@$%^-_=+[{}]:,./?))'. The password cannot contain the
        username or the username in reverse. The Windows ECS password cannot contain the username, the username in reverse,
        or more than two consecutive characters in the username.
    type: str
  data_volumes:
    description:
      - Specifies the data disks of ECS instance.
    type: list
    elements: dict
    suboptions:
      volume_id:
        description:
          - Specifies the disk ID.
        type: str
        required: true
      device:
        description:
          - Specifies the disk device name.
        type: str
  description:
    description:
      - Specifies the description of an ECS, which is a null string by default. Can contain a maximum of 85 characters. Cannot
        contain special characters, such as V(<) and V(>).
    type: str
  eip_id:
    description:
      - Specifies the ID of the elastic IP address assigned to the ECS. Only elastic IP addresses in the DOWN state can be
        assigned.
    type: str
  enable_auto_recovery:
    description:
      - Specifies whether automatic recovery is enabled on the ECS.
    type: bool
  enterprise_project_id:
    description:
      - Specifies the ID of the enterprise project to which the ECS belongs.
    type: str
  security_groups:
    description:
      - Specifies the security groups of the ECS. If this parameter is left blank, the default security group is bound to
        the ECS by default.
    type: list
    elements: str
  server_metadata:
    description:
      - Specifies the metadata of ECS to be created.
    type: dict
  server_tags:
    description:
      - Specifies the tags of an ECS. When you create ECSs, one ECS supports up to 10 tags.
    type: dict
  ssh_key_name:
    description:
      - Specifies the name of the SSH key used for logging in to the ECS.
    type: str
  user_data:
    description:
      - Specifies the user data to be injected during the ECS creation process. Text, text files, and gzip files can be injected.
        The content to be injected must be encoded with base64. The maximum size of the content to be injected (before encoding)
        is 32 KB. For Linux ECSs, this parameter does not take effect when adminPass is used.
    type: str
extends_documentation_fragment:
  - community.general._hwc
  - community.general._attributes
"""

EXAMPLES = r"""
# create an ecs instance
- name: Create a vpc
  hwc_network_vpc:
    cidr: "192.168.100.0/24"
    name: "ansible_network_vpc_test"
  register: vpc
- name: Create a subnet
  hwc_vpc_subnet:
    gateway_ip: "192.168.100.32"
    name: "ansible_network_subnet_test"
    dhcp_enable: true
    vpc_id: "{{ vpc.id }}"
    cidr: "192.168.100.0/26"
  register: subnet
- name: Create a eip
  hwc_vpc_eip:
    dedicated_bandwidth:
      charge_mode: "traffic"
      name: "ansible_test_dedicated_bandwidth"
      size: 1
    type: "5_bgp"
  register: eip
- name: Create a disk
  hwc_evs_disk:
    availability_zone: "cn-north-1a"
    name: "ansible_evs_disk_test"
    volume_type: "SATA"
    size: 10
  register: disk
- name: Create an instance
  community.general.hwc_ecs_instance:
    data_volumes:
      - volume_id: "{{ disk.id }}"
    enable_auto_recovery: false
    eip_id: "{{ eip.id }}"
    name: "ansible_ecs_instance_test"
    availability_zone: "cn-north-1a"
    nics:
      - subnet_id: "{{ subnet.id }}"
        ip_address: "192.168.100.33"
      - subnet_id: "{{ subnet.id }}"
        ip_address: "192.168.100.34"
    server_tags:
      my_server: "my_server"
    image_id: "8da46d6d-6079-4e31-ad6d-a7167efff892"
    flavor_name: "s3.small.1"
    vpc_id: "{{ vpc.id }}"
    root_volume:
      volume_type: "SAS"
"""

RETURN = r"""
availability_zone:
  description:
    - Specifies the name of the AZ where the ECS is located.
  type: str
  returned: success
flavor_name:
  description:
    - Specifies the name of the system flavor.
  type: str
  returned: success
image_id:
  description:
    - Specifies the ID of the system image.
  type: str
  returned: success
name:
  description:
    - Specifies the ECS name. Value requirements "Consists of 1 to 64 characters, including letters, digits, underscores (V(_)),
      hyphens (V(-)), periods (V(.)).".
  type: str
  returned: success
nics:
  description:
    - Specifies the NIC information of the ECS. The network of the NIC must belong to the VPC specified by vpc_id. A maximum
      of 12 NICs can be attached to an ECS.
  type: list
  returned: success
  contains:
    ip_address:
      description:
        - Specifies the IP address of the NIC. The value is an IPv4 address. Its value must be an unused IP address in the
          network segment of the subnet.
      type: str
      returned: success
    subnet_id:
      description:
        - Specifies the ID of subnet.
      type: str
      returned: success
    port_id:
      description:
        - Specifies the port ID corresponding to the IP address.
      type: str
      returned: success
root_volume:
  description:
    - Specifies the configuration of the ECS's system disks.
  type: dict
  returned: success
  contains:
    volume_type:
      description:
        - Specifies the ECS system disk type.
        - SATA is common I/O disk type.
        - SAS is high I/O disk type.
        - SSD is ultra-high I/O disk type.
        - Co-p1 is high I/O (performance-optimized I) disk type.
        - Uh-l1 is ultra-high I/O (latency-optimized) disk type.
        - NOTE is For HANA, HL1, and HL2 ECSs, use co-p1 and uh-l1 disks. For other ECSs, do not use co-p1 or uh-l1 disks.
      type: str
      returned: success
    size:
      description:
        - Specifies the system disk size, in GB. The value range is 1 to 1024. The system disk size must be greater than or
          equal to the minimum system disk size supported by the image (min_disk attribute of the image). If this parameter
          is not specified or is set to 0, the default system disk size is the minimum value of the system disk in the image
          (min_disk attribute of the image).
      type: int
      returned: success
    snapshot_id:
      description:
        - Specifies the snapshot ID or ID of the original data disk contained in the full-ECS image.
      type: str
      returned: success
    device:
      description:
        - Specifies the disk device name.
      type: str
      returned: success
    volume_id:
      description:
        - Specifies the disk ID.
      type: str
      returned: success
vpc_id:
  description:
    - Specifies the ID of the VPC to which the ECS belongs.
  type: str
  returned: success
admin_pass:
  description:
    - Specifies the initial login password of the administrator account for logging in to an ECS using password authentication.
      The Linux administrator is root, and the Windows administrator is Administrator. Password complexity requirements consists
      of 8 to 26 characters. The password must contain at least three of the following character types "uppercase letters,
      lowercase letters, digits, and special characters (!@$%^-_=+[{}]:,./?)". The password cannot contain the username or
      the username in reverse. The Windows ECS password cannot contain the username, the username in reverse, or more than
      two consecutive characters in the username.
  type: str
  returned: success
data_volumes:
  description:
    - Specifies the data disks of ECS instance.
  type: list
  returned: success
  contains:
    volume_id:
      description:
        - Specifies the disk ID.
      type: str
      returned: success
    device:
      description:
        - Specifies the disk device name.
      type: str
      returned: success
description:
  description:
    - Specifies the description of an ECS, which is a null string by default. Can contain a maximum of 85 characters. Cannot
      contain special characters, such as < and >.
  type: str
  returned: success
eip_id:
  description:
    - Specifies the ID of the elastic IP address assigned to the ECS. Only elastic IP addresses in the DOWN state can be assigned.
  type: str
  returned: success
enable_auto_recovery:
  description:
    - Specifies whether automatic recovery is enabled on the ECS.
  type: bool
  returned: success
enterprise_project_id:
  description:
    - Specifies the ID of the enterprise project to which the ECS belongs.
  type: str
  returned: success
security_groups:
  description:
    - Specifies the security groups of the ECS. If this parameter is left blank, the default security group is bound to the
      ECS by default.
  type: list
  returned: success
server_metadata:
  description:
    - Specifies the metadata of ECS to be created.
  type: dict
  returned: success
server_tags:
  description:
    - Specifies the tags of an ECS. When you create ECSs, one ECS supports up to 10 tags.
  type: dict
  returned: success
ssh_key_name:
  description:
    - Specifies the name of the SSH key used for logging in to the ECS.
  type: str
  returned: success
user_data:
  description:
    - Specifies the user data to be injected during the ECS creation process. Text, text files, and gzip files can be injected.
      The content to be injected must be encoded with base64. The maximum size of the content to be injected (before encoding)
      is 32 KB. For Linux ECSs, this parameter does not take effect when adminPass is used.
  type: str
  returned: success
config_drive:
  description:
    - Specifies the configuration driver.
  type: str
  returned: success
created:
  description:
    - Specifies the time when an ECS was created.
  type: str
  returned: success
disk_config_type:
  description:
    - Specifies the disk configuration type. MANUAL is The image space is not expanded. AUTO is the image space of the system
      disk is expanded to be as same as the flavor.
  type: str
  returned: success
host_name:
  description:
    - Specifies the host name of the ECS.
  type: str
  returned: success
image_name:
  description:
    - Specifies the image name of the ECS.
  type: str
  returned: success
power_state:
  description:
    - Specifies the power status of the ECS.
  type: int
  returned: success
server_alias:
  description:
    - Specifies the ECS alias.
  type: str
  returned: success
status:
  description:
    - Specifies the ECS status. Options are ACTIVE, REBOOT, HARD_REBOOT, REBUILD, MIGRATING, BUILD, SHUTOFF, RESIZE, VERIFY_RESIZE,
      ERROR, and DELETED.
  type: str
  returned: success
"""

from ansible_collections.community.general.plugins.module_utils._hwc_utils import (
    Config,
    HwcClientException,
    HwcModule,
    are_different_dicts,
    build_path,
    get_region,
    is_empty_value,
    navigate_value,
    wait_to_finish,
)


def build_module():
    return HwcModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"], type="str"),
            timeouts=dict(
                type="dict",
                options=dict(
                    create=dict(default="30m", type="str"),
                    update=dict(default="30m", type="str"),
                    delete=dict(default="30m", type="str"),
                ),
                default=dict(),
            ),
            availability_zone=dict(type="str", required=True),
            flavor_name=dict(type="str", required=True),
            image_id=dict(type="str", required=True),
            name=dict(type="str", required=True),
            nics=dict(
                type="list",
                required=True,
                elements="dict",
                options=dict(ip_address=dict(type="str", required=True), subnet_id=dict(type="str", required=True)),
            ),
            root_volume=dict(
                type="dict",
                required=True,
                options=dict(
                    volume_type=dict(type="str", required=True), size=dict(type="int"), snapshot_id=dict(type="str")
                ),
            ),
            vpc_id=dict(type="str", required=True),
            admin_pass=dict(type="str", no_log=True),
            data_volumes=dict(
                type="list",
                elements="dict",
                options=dict(volume_id=dict(type="str", required=True), device=dict(type="str")),
            ),
            description=dict(type="str"),
            eip_id=dict(type="str"),
            enable_auto_recovery=dict(type="bool"),
            enterprise_project_id=dict(type="str"),
            security_groups=dict(type="list", elements="str"),
            server_metadata=dict(type="dict"),
            server_tags=dict(type="dict"),
            ssh_key_name=dict(type="str"),
            user_data=dict(type="str"),
        ),
        supports_check_mode=True,
    )


def main():
    """Main function"""

    module = build_module()
    config = Config(module, "ecs")

    try:
        _init(config)
        is_exist = module.params["id"]

        result = None
        changed = False
        if module.params["state"] == "present":
            if not is_exist:
                if not module.check_mode:
                    create(config)
                changed = True

            inputv = user_input_parameters(module)
            resp, array_index = read_resource(config)
            result = build_state(inputv, resp, array_index)
            set_readonly_options(inputv, result)
            if are_different_dicts(inputv, result):
                if not module.check_mode:
                    update(config, inputv, result)

                    inputv = user_input_parameters(module)
                    resp, array_index = read_resource(config)
                    result = build_state(inputv, resp, array_index)
                    set_readonly_options(inputv, result)
                    if are_different_dicts(inputv, result):
                        raise Exception("Update resource failed, some attributes are not updated")

                changed = True

            result["id"] = module.params.get("id")
        else:
            result = dict()
            if is_exist:
                if not module.check_mode:
                    delete(config)
                changed = True

    except Exception as ex:
        module.fail_json(msg=str(ex))

    else:
        result["changed"] = changed
        module.exit_json(**result)


def _init(config):
    module = config.module
    if module.params["id"]:
        return

    v = search_resource(config)
    n = len(v)
    if n > 1:
        raise Exception(f"Found more than one resource({', '.join([navigate_value(i, ['id']) for i in v])})")

    if n == 1:
        module.params["id"] = navigate_value(v[0], ["id"])


def user_input_parameters(module):
    return {
        "admin_pass": module.params.get("admin_pass"),
        "availability_zone": module.params.get("availability_zone"),
        "data_volumes": module.params.get("data_volumes"),
        "description": module.params.get("description"),
        "eip_id": module.params.get("eip_id"),
        "enable_auto_recovery": module.params.get("enable_auto_recovery"),
        "enterprise_project_id": module.params.get("enterprise_project_id"),
        "flavor_name": module.params.get("flavor_name"),
        "image_id": module.params.get("image_id"),
        "name": module.params.get("name"),
        "nics": module.params.get("nics"),
        "root_volume": module.params.get("root_volume"),
        "security_groups": module.params.get("security_groups"),
        "server_metadata": module.params.get("server_metadata"),
        "server_tags": module.params.get("server_tags"),
        "ssh_key_name": module.params.get("ssh_key_name"),
        "user_data": module.params.get("user_data"),
        "vpc_id": module.params.get("vpc_id"),
    }


def create(config):
    module = config.module
    client = config.client(get_region(module), "ecs", "project")
    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    opts = user_input_parameters(module)
    opts["ansible_module"] = module

    params = build_create_parameters(opts)
    r = send_create_request(module, params, client)
    obj = async_wait(config, r, client, timeout)

    sub_job_identity = {
        "job_type": "createSingleServer",
    }
    for item in navigate_value(obj, ["entities", "sub_jobs"]):
        for k, v in sub_job_identity.items():
            if item[k] != v:
                break
        else:
            obj = item
            break
    else:
        raise Exception("Can't find the sub job")
    module.params["id"] = navigate_value(obj, ["entities", "server_id"])


def update(config, expect_state, current_state):
    module = config.module
    expect_state["current_state"] = current_state
    current_state["current_state"] = current_state
    timeout = 60 * int(module.params["timeouts"]["update"].rstrip("m"))
    client = config.client(get_region(module), "ecs", "project")

    params = build_delete_nics_parameters(expect_state)
    params1 = build_delete_nics_parameters(current_state)
    if params and are_different_dicts(params, params1):
        r = send_delete_nics_request(module, params, client)
        async_wait(config, r, client, timeout)

    params = build_set_auto_recovery_parameters(expect_state)
    params1 = build_set_auto_recovery_parameters(current_state)
    if params and are_different_dicts(params, params1):
        send_set_auto_recovery_request(module, params, client)

    params = build_attach_nics_parameters(expect_state)
    params1 = build_attach_nics_parameters(current_state)
    if params and are_different_dicts(params, params1):
        r = send_attach_nics_request(module, params, client)
        async_wait(config, r, client, timeout)

    multi_invoke_delete_volume(config, expect_state, client, timeout)

    multi_invoke_attach_data_disk(config, expect_state, client, timeout)


def delete(config):
    module = config.module
    client = config.client(get_region(module), "ecs", "project")
    timeout = 60 * int(module.params["timeouts"]["delete"].rstrip("m"))

    opts = user_input_parameters(module)
    opts["ansible_module"] = module

    params = build_delete_parameters(opts)
    if params:
        r = send_delete_request(module, params, client)
        async_wait(config, r, client, timeout)


def read_resource(config):
    module = config.module
    client = config.client(get_region(module), "ecs", "project")

    res = {}

    r = send_read_request(module, client)
    preprocess_read_response(r)
    res["read"] = fill_read_resp_body(r)

    r = send_read_auto_recovery_request(module, client)
    res["read_auto_recovery"] = fill_read_auto_recovery_resp_body(r)

    return res, None


def preprocess_read_response(resp):
    v = resp.get("os-extended-volumes:volumes_attached")
    if v and isinstance(v, list):
        for i in range(len(v)):
            if v[i].get("bootIndex") == "0":
                root_volume = v[i]

                if (i + 1) != len(v):
                    v[i] = v[-1]

                v.pop()

                resp["root_volume"] = root_volume
                break

    v = resp.get("addresses")
    if v:
        rv = {}
        eips = []
        for val in v.values():
            for item in val:
                if item["OS-EXT-IPS:type"] == "floating":
                    eips.append(item)
                else:
                    rv[item["OS-EXT-IPS:port_id"]] = item

        for item in eips:
            k = item["OS-EXT-IPS:port_id"]
            if k in rv:
                rv[k]["eip_address"] = item.get("addr", "")
            else:
                rv[k] = item
                item["eip_address"] = item.get("addr", "")
                item["addr"] = ""

        resp["address"] = rv.values()


def build_state(opts, response, array_index):
    states = flatten_options(response, array_index)
    set_unreadable_options(opts, states)
    adjust_options(opts, states)
    return states


def _build_query_link(opts):
    query_params = []

    v = navigate_value(opts, ["enterprise_project_id"])
    if v or v in [False, 0]:
        query_params.append(f"enterprise_project_id={str(v) if v else str(v).lower()}")

    v = navigate_value(opts, ["name"])
    if v or v in [False, 0]:
        query_params.append(f"name={str(v) if v else str(v).lower()}")

    query_link = "?limit=10&offset={offset}"
    if query_params:
        query_link += f"&{'&'.join(query_params)}"

    return query_link


def search_resource(config):
    module = config.module
    client = config.client(get_region(module), "ecs", "project")
    opts = user_input_parameters(module)
    identity_obj = _build_identity_object(opts)
    query_link = _build_query_link(opts)
    link = f"cloudservers/detail{query_link}"

    result = []
    p = {"offset": 1}
    while True:
        url = link.format(**p)
        r = send_list_request(module, client, url)
        if not r:
            break

        for item in r:
            item = fill_list_resp_body(item)
            adjust_list_resp(identity_obj, item)
            if not are_different_dicts(identity_obj, item):
                result.append(item)

        if len(result) > 1:
            break

        p["offset"] += 1

    return result


def build_delete_nics_parameters(opts):
    params = dict()

    v = expand_delete_nics_nics(opts, None)
    if not is_empty_value(v):
        params["nics"] = v

    return params


def expand_delete_nics_nics(d, array_index):
    cv = d["current_state"].get("nics")
    if not cv:
        return None

    val = cv

    ev = d.get("nics")
    if ev:
        m = [item.get("ip_address") for item in ev]
        val = [item for item in cv if item.get("ip_address") not in m]

    r = []
    for item in val:
        transformed = dict()

        v = item.get("port_id")
        if not is_empty_value(v):
            transformed["id"] = v

        if transformed:
            r.append(transformed)

    return r


def send_delete_nics_request(module, params, client):
    url = build_path(module, "cloudservers/{id}/nics/delete")

    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(delete_nics), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_set_auto_recovery_parameters(opts):
    params = dict()

    v = expand_set_auto_recovery_support_auto_recovery(opts, None)
    if v is not None:
        params["support_auto_recovery"] = v

    return params


def expand_set_auto_recovery_support_auto_recovery(d, array_index):
    v = navigate_value(d, ["enable_auto_recovery"], None)
    return None if v is None else str(v).lower()


def send_set_auto_recovery_request(module, params, client):
    url = build_path(module, "cloudservers/{id}/autorecovery")

    try:
        r = client.put(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(set_auto_recovery), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_create_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["admin_pass"], None)
    if not is_empty_value(v):
        params["adminPass"] = v

    v = navigate_value(opts, ["availability_zone"], None)
    if not is_empty_value(v):
        params["availability_zone"] = v

    v = navigate_value(opts, ["description"], None)
    if not is_empty_value(v):
        params["description"] = v

    v = expand_create_extendparam(opts, None)
    if not is_empty_value(v):
        params["extendparam"] = v

    v = navigate_value(opts, ["flavor_name"], None)
    if not is_empty_value(v):
        params["flavorRef"] = v

    v = navigate_value(opts, ["image_id"], None)
    if not is_empty_value(v):
        params["imageRef"] = v

    v = navigate_value(opts, ["ssh_key_name"], None)
    if not is_empty_value(v):
        params["key_name"] = v

    v = navigate_value(opts, ["server_metadata"], None)
    if not is_empty_value(v):
        params["metadata"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = expand_create_nics(opts, None)
    if not is_empty_value(v):
        params["nics"] = v

    v = expand_create_publicip(opts, None)
    if not is_empty_value(v):
        params["publicip"] = v

    v = expand_create_root_volume(opts, None)
    if not is_empty_value(v):
        params["root_volume"] = v

    v = expand_create_security_groups(opts, None)
    if not is_empty_value(v):
        params["security_groups"] = v

    v = expand_create_server_tags(opts, None)
    if not is_empty_value(v):
        params["server_tags"] = v

    v = navigate_value(opts, ["user_data"], None)
    if not is_empty_value(v):
        params["user_data"] = v

    v = navigate_value(opts, ["vpc_id"], None)
    if not is_empty_value(v):
        params["vpcid"] = v

    if not params:
        return params

    params = {"server": params}

    return params


def expand_create_extendparam(d, array_index):
    r = dict()

    r["chargingMode"] = 0

    v = navigate_value(d, ["enterprise_project_id"], array_index)
    if not is_empty_value(v):
        r["enterprise_project_id"] = v

    v = navigate_value(d, ["enable_auto_recovery"], array_index)
    if not is_empty_value(v):
        r["support_auto_recovery"] = v

    return r


def expand_create_nics(d, array_index):
    new_ai = dict()
    if array_index:
        new_ai.update(array_index)

    req = []

    v = navigate_value(d, ["nics"], new_ai)

    if not v:
        return req
    n = len(v)
    for i in range(n):
        new_ai["nics"] = i
        transformed = dict()

        v = navigate_value(d, ["nics", "ip_address"], new_ai)
        if not is_empty_value(v):
            transformed["ip_address"] = v

        v = navigate_value(d, ["nics", "subnet_id"], new_ai)
        if not is_empty_value(v):
            transformed["subnet_id"] = v

        if transformed:
            req.append(transformed)

    return req


def expand_create_publicip(d, array_index):
    r = dict()

    v = navigate_value(d, ["eip_id"], array_index)
    if not is_empty_value(v):
        r["id"] = v

    return r


def expand_create_root_volume(d, array_index):
    r = dict()

    v = expand_create_root_volume_extendparam(d, array_index)
    if not is_empty_value(v):
        r["extendparam"] = v

    v = navigate_value(d, ["root_volume", "size"], array_index)
    if not is_empty_value(v):
        r["size"] = v

    v = navigate_value(d, ["root_volume", "volume_type"], array_index)
    if not is_empty_value(v):
        r["volumetype"] = v

    return r


def expand_create_root_volume_extendparam(d, array_index):
    r = dict()

    v = navigate_value(d, ["root_volume", "snapshot_id"], array_index)
    if not is_empty_value(v):
        r["snapshotId"] = v

    return r


def expand_create_security_groups(d, array_index):
    v = d.get("security_groups")
    if not v:
        return None

    return [{"id": i} for i in v]


def expand_create_server_tags(d, array_index):
    v = d.get("server_tags")
    if not v:
        return None

    return [{"key": k, "value": v1} for k, v1 in v.items()]


def send_create_request(module, params, client):
    url = "cloudservers"
    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(create), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_attach_nics_parameters(opts):
    params = dict()

    v = expand_attach_nics_nics(opts, None)
    if not is_empty_value(v):
        params["nics"] = v

    return params


def expand_attach_nics_nics(d, array_index):
    ev = d.get("nics")
    if not ev:
        return None

    val = ev

    cv = d["current_state"].get("nics")
    if cv:
        m = [item.get("ip_address") for item in cv]
        val = [item for item in ev if item.get("ip_address") not in m]

    r = []
    for item in val:
        transformed = dict()

        v = item.get("ip_address")
        if not is_empty_value(v):
            transformed["ip_address"] = v

        v = item.get("subnet_id")
        if not is_empty_value(v):
            transformed["subnet_id"] = v

        if transformed:
            r.append(transformed)

    return r


def send_attach_nics_request(module, params, client):
    url = build_path(module, "cloudservers/{id}/nics")

    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(attach_nics), error: {ex}"
        module.fail_json(msg=msg)

    return r


def send_delete_volume_request(module, params, client, info):
    path_parameters = {
        "volume_id": ["volume_id"],
    }
    data = {key: navigate_value(info, path) for key, path in path_parameters.items()}

    url = build_path(module, "cloudservers/{id}/detachvolume/{volume_id}", data)

    try:
        r = client.delete(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(delete_volume), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_attach_data_disk_parameters(opts, array_index):
    params = dict()

    v = expand_attach_data_disk_volume_attachment(opts, array_index)
    if not is_empty_value(v):
        params["volumeAttachment"] = v

    return params


def expand_attach_data_disk_volume_attachment(d, array_index):
    r = dict()

    v = navigate_value(d, ["data_volumes", "device"], array_index)
    if not is_empty_value(v):
        r["device"] = v

    v = navigate_value(d, ["data_volumes", "volume_id"], array_index)
    if not is_empty_value(v):
        r["volumeId"] = v

    return r


def send_attach_data_disk_request(module, params, client):
    url = build_path(module, "cloudservers/{id}/attachvolume")

    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(attach_data_disk), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_delete_parameters(opts):
    params = dict()

    params["delete_publicip"] = False

    params["delete_volume"] = False

    v = expand_delete_servers(opts, None)
    if not is_empty_value(v):
        params["servers"] = v

    return params


def expand_delete_servers(d, array_index):
    new_ai = dict()
    if array_index:
        new_ai.update(array_index)

    req = []

    transformed = dict()

    v = expand_delete_servers_id(d, new_ai)
    if not is_empty_value(v):
        transformed["id"] = v

    if transformed:
        req.append(transformed)

    return req


def expand_delete_servers_id(d, array_index):
    return d["ansible_module"].params.get("id")


def send_delete_request(module, params, client):
    url = "cloudservers/delete"
    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(delete), error: {ex}"
        module.fail_json(msg=msg)

    return r


def async_wait(config, result, client, timeout):
    module = config.module

    url = build_path(module, "jobs/{job_id}", result)

    def _query_status():
        r = None
        try:
            r = client.get(url, timeout=timeout)
        except HwcClientException:
            return None, ""

        try:
            s = navigate_value(r, ["status"])
            return r, s
        except Exception:
            return None, ""

    try:
        return wait_to_finish(["SUCCESS"], ["RUNNING", "INIT"], _query_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_ecs_instance): error waiting to be done, error= {ex}")


def multi_invoke_delete_volume(config, opts, client, timeout):
    module = config.module

    opts1 = None
    expect = opts["data_volumes"]
    current = opts["current_state"]["data_volumes"]
    if expect and current:
        v = [i["volume_id"] for i in expect]
        opts1 = {"data_volumes": [i for i in current if i["volume_id"] not in v]}

    loop_val = navigate_value(opts1, ["data_volumes"])
    if not loop_val:
        return

    for i in range(len(loop_val)):
        r = send_delete_volume_request(module, None, client, loop_val[i])
        async_wait(config, r, client, timeout)


def multi_invoke_attach_data_disk(config, opts, client, timeout):
    module = config.module

    opts1 = opts
    expect = opts["data_volumes"]
    current = opts["current_state"]["data_volumes"]
    if expect and current:
        v = [i["volume_id"] for i in current]
        opts1 = {"data_volumes": [i for i in expect if i["volume_id"] not in v]}

    loop_val = navigate_value(opts1, ["data_volumes"])
    if not loop_val:
        return

    for i in range(len(loop_val)):
        params = build_attach_data_disk_parameters(opts1, {"data_volumes": i})
        r = send_attach_data_disk_request(module, params, client)
        async_wait(config, r, client, timeout)


def send_read_request(module, client):
    url = build_path(module, "cloudservers/{id}")

    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(read), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["server"], None)


def fill_read_resp_body(body):
    result = dict()

    result["OS-DCF:diskConfig"] = body.get("OS-DCF:diskConfig")

    result["OS-EXT-AZ:availability_zone"] = body.get("OS-EXT-AZ:availability_zone")

    result["OS-EXT-SRV-ATTR:hostname"] = body.get("OS-EXT-SRV-ATTR:hostname")

    result["OS-EXT-SRV-ATTR:instance_name"] = body.get("OS-EXT-SRV-ATTR:instance_name")

    result["OS-EXT-SRV-ATTR:user_data"] = body.get("OS-EXT-SRV-ATTR:user_data")

    result["OS-EXT-STS:power_state"] = body.get("OS-EXT-STS:power_state")

    v = fill_read_resp_address(body.get("address"))
    result["address"] = v

    result["config_drive"] = body.get("config_drive")

    result["created"] = body.get("created")

    result["description"] = body.get("description")

    result["enterprise_project_id"] = body.get("enterprise_project_id")

    v = fill_read_resp_flavor(body.get("flavor"))
    result["flavor"] = v

    result["id"] = body.get("id")

    v = fill_read_resp_image(body.get("image"))
    result["image"] = v

    result["key_name"] = body.get("key_name")

    v = fill_read_resp_metadata(body.get("metadata"))
    result["metadata"] = v

    result["name"] = body.get("name")

    v = fill_read_resp_os_extended_volumes_volumes_attached(body.get("os-extended-volumes:volumes_attached"))
    result["os-extended-volumes:volumes_attached"] = v

    v = fill_read_resp_root_volume(body.get("root_volume"))
    result["root_volume"] = v

    result["status"] = body.get("status")

    result["tags"] = body.get("tags")

    return result


def fill_read_resp_address(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["OS-EXT-IPS:port_id"] = item.get("OS-EXT-IPS:port_id")

        val["OS-EXT-IPS:type"] = item.get("OS-EXT-IPS:type")

        val["addr"] = item.get("addr")

        result.append(val)

    return result


def fill_read_resp_flavor(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


def fill_read_resp_image(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


def fill_read_resp_metadata(value):
    if not value:
        return None

    result = dict()

    result["image_name"] = value.get("image_name")

    result["vpc_id"] = value.get("vpc_id")

    return result


def fill_read_resp_os_extended_volumes_volumes_attached(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["bootIndex"] = item.get("bootIndex")

        val["device"] = item.get("device")

        val["id"] = item.get("id")

        result.append(val)

    return result


def fill_read_resp_root_volume(value):
    if not value:
        return None

    result = dict()

    result["device"] = value.get("device")

    result["id"] = value.get("id")

    return result


def send_read_auto_recovery_request(module, client):
    url = build_path(module, "cloudservers/{id}/autorecovery")

    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(read_auto_recovery), error: {ex}"
        module.fail_json(msg=msg)

    return r


def fill_read_auto_recovery_resp_body(body):
    result = dict()

    result["support_auto_recovery"] = body.get("support_auto_recovery")

    return result


def flatten_options(response, array_index):
    r = dict()

    v = navigate_value(response, ["read", "OS-EXT-AZ:availability_zone"], array_index)
    r["availability_zone"] = v

    v = navigate_value(response, ["read", "config_drive"], array_index)
    r["config_drive"] = v

    v = navigate_value(response, ["read", "created"], array_index)
    r["created"] = v

    v = flatten_data_volumes(response, array_index)
    r["data_volumes"] = v

    v = navigate_value(response, ["read", "description"], array_index)
    r["description"] = v

    v = navigate_value(response, ["read", "OS-DCF:diskConfig"], array_index)
    r["disk_config_type"] = v

    v = flatten_enable_auto_recovery(response, array_index)
    r["enable_auto_recovery"] = v

    v = navigate_value(response, ["read", "enterprise_project_id"], array_index)
    r["enterprise_project_id"] = v

    v = navigate_value(response, ["read", "flavor", "id"], array_index)
    r["flavor_name"] = v

    v = navigate_value(response, ["read", "OS-EXT-SRV-ATTR:hostname"], array_index)
    r["host_name"] = v

    v = navigate_value(response, ["read", "image", "id"], array_index)
    r["image_id"] = v

    v = navigate_value(response, ["read", "metadata", "image_name"], array_index)
    r["image_name"] = v

    v = navigate_value(response, ["read", "name"], array_index)
    r["name"] = v

    v = flatten_nics(response, array_index)
    r["nics"] = v

    v = navigate_value(response, ["read", "OS-EXT-STS:power_state"], array_index)
    r["power_state"] = v

    v = flatten_root_volume(response, array_index)
    r["root_volume"] = v

    v = navigate_value(response, ["read", "OS-EXT-SRV-ATTR:instance_name"], array_index)
    r["server_alias"] = v

    v = flatten_server_tags(response, array_index)
    r["server_tags"] = v

    v = navigate_value(response, ["read", "key_name"], array_index)
    r["ssh_key_name"] = v

    v = navigate_value(response, ["read", "status"], array_index)
    r["status"] = v

    v = navigate_value(response, ["read", "OS-EXT-SRV-ATTR:user_data"], array_index)
    r["user_data"] = v

    v = navigate_value(response, ["read", "metadata", "vpc_id"], array_index)
    r["vpc_id"] = v

    return r


def flatten_data_volumes(d, array_index):
    v = navigate_value(d, ["read", "os-extended-volumes:volumes_attached"], array_index)
    if not v:
        return None
    n = len(v)
    result = []

    new_ai = dict()
    if array_index:
        new_ai.update(array_index)

    for i in range(n):
        new_ai["read.os-extended-volumes:volumes_attached"] = i

        val = dict()

        v = navigate_value(d, ["read", "os-extended-volumes:volumes_attached", "device"], new_ai)
        val["device"] = v

        v = navigate_value(d, ["read", "os-extended-volumes:volumes_attached", "id"], new_ai)
        val["volume_id"] = v

        for v in val.values():
            if v is not None:
                result.append(val)
                break

    return result if result else None


def flatten_enable_auto_recovery(d, array_index):
    v = navigate_value(d, ["read_auto_recovery", "support_auto_recovery"], array_index)
    return v == "true"


def flatten_nics(d, array_index):
    v = navigate_value(d, ["read", "address"], array_index)
    if not v:
        return None
    n = len(v)
    result = []

    new_ai = dict()
    if array_index:
        new_ai.update(array_index)

    for i in range(n):
        new_ai["read.address"] = i

        val = dict()

        v = navigate_value(d, ["read", "address", "addr"], new_ai)
        val["ip_address"] = v

        v = navigate_value(d, ["read", "address", "OS-EXT-IPS:port_id"], new_ai)
        val["port_id"] = v

        for v in val.values():
            if v is not None:
                result.append(val)
                break

    return result if result else None


def flatten_root_volume(d, array_index):
    result = dict()

    v = navigate_value(d, ["read", "root_volume", "device"], array_index)
    result["device"] = v

    v = navigate_value(d, ["read", "root_volume", "id"], array_index)
    result["volume_id"] = v

    for v in result.values():
        if v is not None:
            return result
    return None


def flatten_server_tags(d, array_index):
    v = navigate_value(d, ["read", "tags"], array_index)
    if not v:
        return None

    r = dict()
    for item in v:
        v1 = item.split("=")
        if v1:
            r[v1[0]] = v1[1]
    return r


def adjust_options(opts, states):
    adjust_data_volumes(opts, states)

    adjust_nics(opts, states)


def adjust_data_volumes(parent_input, parent_cur):
    iv = parent_input.get("data_volumes")
    if not (iv and isinstance(iv, list)):
        return

    cv = parent_cur.get("data_volumes")
    if not (cv and isinstance(cv, list)):
        return

    lcv = len(cv)
    result = []
    q = []
    for iiv in iv:
        if len(q) == lcv:
            break

        icv = None
        for j in range(lcv):
            if j in q:
                continue

            icv = cv[j]

            if iiv["volume_id"] != icv["volume_id"]:
                continue

            result.append(icv)
            q.append(j)
            break
        else:
            break

    if len(q) != lcv:
        for i in range(lcv):
            if i not in q:
                result.append(cv[i])

    if len(result) != lcv:
        raise Exception("adjust property(data_volumes) failed, the array number is not equal")

    parent_cur["data_volumes"] = result


def adjust_nics(parent_input, parent_cur):
    iv = parent_input.get("nics")
    if not (iv and isinstance(iv, list)):
        return

    cv = parent_cur.get("nics")
    if not (cv and isinstance(cv, list)):
        return

    lcv = len(cv)
    result = []
    q = []
    for iiv in iv:
        if len(q) == lcv:
            break

        icv = None
        for j in range(lcv):
            if j in q:
                continue

            icv = cv[j]

            if iiv["ip_address"] != icv["ip_address"]:
                continue

            result.append(icv)
            q.append(j)
            break
        else:
            break

    if len(q) != lcv:
        for i in range(lcv):
            if i not in q:
                result.append(cv[i])

    if len(result) != lcv:
        raise Exception("adjust property(nics) failed, the array number is not equal")

    parent_cur["nics"] = result


def set_unreadable_options(opts, states):
    states["admin_pass"] = opts.get("admin_pass")

    states["eip_id"] = opts.get("eip_id")

    set_unread_nics(opts.get("nics"), states.get("nics"))

    set_unread_root_volume(opts.get("root_volume"), states.get("root_volume"))

    states["security_groups"] = opts.get("security_groups")

    states["server_metadata"] = opts.get("server_metadata")


def set_unread_nics(inputv, curv):
    if not (inputv and isinstance(inputv, list)):
        return

    if not (curv and isinstance(curv, list)):
        return

    lcv = len(curv)
    q = []
    for iv in inputv:
        if len(q) == lcv:
            break

        cv = None
        for j in range(lcv):
            if j in q:
                continue

            cv = curv[j]

            if iv["ip_address"] != cv["ip_address"]:
                continue

            q.append(j)
            break
        else:
            continue

        cv["subnet_id"] = iv.get("subnet_id")


def set_unread_root_volume(inputv, curv):
    if not (inputv and isinstance(inputv, dict)):
        return

    if not (curv and isinstance(curv, dict)):
        return

    curv["size"] = inputv.get("size")

    curv["snapshot_id"] = inputv.get("snapshot_id")

    curv["volume_type"] = inputv.get("volume_type")


def set_readonly_options(opts, states):
    opts["config_drive"] = states.get("config_drive")

    opts["created"] = states.get("created")

    opts["disk_config_type"] = states.get("disk_config_type")

    opts["host_name"] = states.get("host_name")

    opts["image_name"] = states.get("image_name")

    set_readonly_nics(opts.get("nics"), states.get("nics"))

    opts["power_state"] = states.get("power_state")

    set_readonly_root_volume(opts.get("root_volume"), states.get("root_volume"))

    opts["server_alias"] = states.get("server_alias")

    opts["status"] = states.get("status")


def set_readonly_nics(inputv, curv):
    if not (curv and isinstance(curv, list)):
        return

    if not (inputv and isinstance(inputv, list)):
        return

    lcv = len(curv)
    q = []
    for iv in inputv:
        if len(q) == lcv:
            break

        cv = None
        for j in range(lcv):
            if j in q:
                continue

            cv = curv[j]

            if iv["ip_address"] != cv["ip_address"]:
                continue

            q.append(j)
            break
        else:
            continue

        iv["port_id"] = cv.get("port_id")


def set_readonly_root_volume(inputv, curv):
    if not (inputv and isinstance(inputv, dict)):
        return

    if not (curv and isinstance(curv, dict)):
        return

    inputv["device"] = curv.get("device")

    inputv["volume_id"] = curv.get("volume_id")


def send_list_request(module, client, url):
    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_ecs_instance): error running api(list), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["servers"], None)


def _build_identity_object(all_opts):
    result = dict()

    result["OS-DCF:diskConfig"] = None

    v = navigate_value(all_opts, ["availability_zone"], None)
    result["OS-EXT-AZ:availability_zone"] = v

    result["OS-EXT-SRV-ATTR:hostname"] = None

    result["OS-EXT-SRV-ATTR:instance_name"] = None

    v = navigate_value(all_opts, ["user_data"], None)
    result["OS-EXT-SRV-ATTR:user_data"] = v

    result["OS-EXT-STS:power_state"] = None

    result["config_drive"] = None

    result["created"] = None

    v = navigate_value(all_opts, ["description"], None)
    result["description"] = v

    v = navigate_value(all_opts, ["enterprise_project_id"], None)
    result["enterprise_project_id"] = v

    v = expand_list_flavor(all_opts, None)
    result["flavor"] = v

    result["id"] = None

    v = expand_list_image(all_opts, None)
    result["image"] = v

    v = navigate_value(all_opts, ["ssh_key_name"], None)
    result["key_name"] = v

    v = expand_list_metadata(all_opts, None)
    result["metadata"] = v

    v = navigate_value(all_opts, ["name"], None)
    result["name"] = v

    result["status"] = None

    v = expand_list_tags(all_opts, None)
    result["tags"] = v

    return result


def expand_list_flavor(d, array_index):
    r = dict()

    v = navigate_value(d, ["flavor_name"], array_index)
    r["id"] = v

    for v in r.values():
        if v is not None:
            return r
    return None


def expand_list_image(d, array_index):
    r = dict()

    v = navigate_value(d, ["image_id"], array_index)
    r["id"] = v

    for v in r.values():
        if v is not None:
            return r
    return None


def expand_list_metadata(d, array_index):
    r = dict()

    v = navigate_value(d, ["vpc_id"], array_index)
    r["vpc_id"] = v

    for v in r.values():
        if v is not None:
            return r
    return None


def expand_list_tags(d, array_index):
    v = d.get("server_tags")
    if not v:
        return None

    return [f"{k}={v1}" for k, v1 in v.items()]


def fill_list_resp_body(body):
    result = dict()

    result["OS-DCF:diskConfig"] = body.get("OS-DCF:diskConfig")

    result["OS-EXT-AZ:availability_zone"] = body.get("OS-EXT-AZ:availability_zone")

    result["OS-EXT-SRV-ATTR:hostname"] = body.get("OS-EXT-SRV-ATTR:hostname")

    result["OS-EXT-SRV-ATTR:instance_name"] = body.get("OS-EXT-SRV-ATTR:instance_name")

    result["OS-EXT-SRV-ATTR:user_data"] = body.get("OS-EXT-SRV-ATTR:user_data")

    result["OS-EXT-STS:power_state"] = body.get("OS-EXT-STS:power_state")

    result["config_drive"] = body.get("config_drive")

    result["created"] = body.get("created")

    result["description"] = body.get("description")

    result["enterprise_project_id"] = body.get("enterprise_project_id")

    v = fill_list_resp_flavor(body.get("flavor"))
    result["flavor"] = v

    result["id"] = body.get("id")

    v = fill_list_resp_image(body.get("image"))
    result["image"] = v

    result["key_name"] = body.get("key_name")

    v = fill_list_resp_metadata(body.get("metadata"))
    result["metadata"] = v

    result["name"] = body.get("name")

    result["status"] = body.get("status")

    result["tags"] = body.get("tags")

    return result


def fill_list_resp_flavor(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


def fill_list_resp_image(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


def fill_list_resp_metadata(value):
    if not value:
        return None

    result = dict()

    result["vpc_id"] = value.get("vpc_id")

    return result


def adjust_list_resp(opts, resp):
    adjust_list_api_tags(opts, resp)


def adjust_list_api_tags(parent_input, parent_cur):
    iv = parent_input.get("tags")
    if not (iv and isinstance(iv, list)):
        return

    cv = parent_cur.get("tags")
    if not (cv and isinstance(cv, list)):
        return

    result = []
    for iiv in iv:
        if iiv not in cv:
            break

        result.append(iiv)

        j = cv.index(iiv)
        cv[j] = cv[-1]
        cv.pop()

    if cv:
        result.extend(cv)
    parent_cur["tags"] = result


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_appliance_firewall_firewalled_services
short_description: Resource module for networks _appliance _firewall _firewalled _services
description:
  - Manage operation update of the resource networks _appliance _firewall _firewalled
    _services.
  - Updates the accessibility settings for the given service 'ICMP', 'web', or 'SNMP'
    .
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  access:
    description: A string indicating the rule for which IPs are allowed to use the
      specified service. Acceptable values are "blocked" (no remote IPs can access
      the service), "restricted" (only allowed IPs can access the service), and "unrestriced"
      (any remote IP can access the service). This field is required.
    type: str
  allowedIps:
    description: An array of allowed CIDRs that can access the service. This field
      is required if "access" is set to "restricted". Otherwise this field is ignored.
    elements: str
    type: list
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  service:
    description: Service path parameter.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for appliance updateNetworkApplianceFirewallFirewalledService
    description: Complete reference of the updateNetworkApplianceFirewallFirewalledService
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-appliance-firewall-firewalled-service
notes:
  - SDK Method used are
    appliance.Appliance.update_network_appliance_firewall_firewalled_service,
  - Paths used are
    put /networks/{networkId}/appliance/firewall/firewalledServices/{service},
"""

EXAMPLES = r"""
- name: Update by id
  cisco.meraki.networks_appliance_firewall_firewalled_services:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    access: restricted
    allowedIps:
      - 123.123.123.1
    networkId: string
    service: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "service": "string",
      "access": "string",
      "allowedIps": [
        "string"
      ]
    }
"""



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Use AnsibleModule's argspec validation

def _check_argspec(self):
    aav = AnsibleArgSpecValidator(
        data=self._task.args,
        schema=DOCUMENTATION,
        schema_format="doc",
        schema_conditionals={},
        other_args={},
        name=self._task.action,
    )
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleActionFail(errors)

"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.ansible.utils.plugins.module_utils.common.utils import dict_merge


try:
    import yaml

    # use C version if possible for speedup
    try:
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader
    HAS_YAML = True
except ImportError:
    HAS_YAML = False

# TODO: Update this to point to functionality being exposed in 2.11
# ansible-base 2.11 should expose argspec validation outside of the
# ansiblemodule class
try:
    from ansible.module_utils.common.arg_spec import ArgumentSpecValidator

    HAS_ANSIBLE_ARG_SPEC_VALIDATOR = True
except ImportError:
    HAS_ANSIBLE_ARG_SPEC_VALIDATOR = False


OPTION_METADATA = (
    "type",
    "choices",
    "default",
    "required",
    "aliases",
    "elements",
    "fallback",
    "no_log",
    "apply_defaults",
    "deprecated_aliases",
    "removed_in_version",
)
OPTION_CONDITIONALS = (
    "mutually_exclusive",
    "required_one_of",
    "required_together",
    "required_by",
    "required_if",
)

VALID_ANSIBLEMODULE_ARGS = (
    "argument_spec",
    "bypass_checks",
    "no_log",
    "add_file_common_args",
    "supports_check_mode",
) + OPTION_CONDITIONALS

BASE_ARG_AVAIL = 2.11


class MonkeyModule(AnsibleModule):
    """A derivative of the AnsibleModule used
    to just validate the data (task.args) against
    the schema(argspec)
    """

    def __init__(self, data, schema, name):
        self._errors = None
        self._valid = True
        self._schema = schema
        self.name = name
        self.params = data

    def fail_json(self, msg):
        """Replace the AnsibleModule fail_json here
        :param msg: The message for the failure
        :type msg: str
        """
        if self.name:
            msg = re.sub(r"\(basic\.pyc?\)", "'{name}'".format(name=self.name), msg)
        self._valid = False
        self._errors = msg

    def _load_params(self):
        """This replaces the AnsibleModule _load_params
        fn because we already set self.params in init
        """
        pass

    def validate(self):
        """Instantiate the super, validating the schema
        against the data
        :return valid: if the data passed
        :rtype valid: bool
        :return errors: errors reported during validation
        :rtype errors: str
        :return params: The original data updated with defaults
        :rtype params: dict
        """
        super(MonkeyModule, self).__init__(**self._schema)
        return self._valid, self._errors, self.params


class AnsibleArgSpecValidator:
    def __init__(
        self,
        data,
        schema,
        schema_format="doc",
        schema_conditionals=None,
        name=None,
        other_args=None,
    ):
        """Validate some data against a schema
        :param data: The data to validate
        :type data: dict
        :param schema: A schema in ansible argspec format
        :type schema: dict
        :param schema_format: 'doc' (ansible docstring) or 'argspec' (ansible argspec)
        :type schema: str if doc, dict if argspec
        :param schema_conditionals: A dict of schema conditionals, ie required_if
        :type schema_conditionals: dict
        :param name: the name of the plugin calling this class, used in error messages
        :type name: str
        :param other_args: Other valid kv pairs for the argspec, eg no_log, bypass_checks
        :type other_args: dict

        note:
        - the schema conditionals can be root conditionals or deeply nested conditionals
          these get dict_merged into the argspec from the docstring, since the docstring cannot
          contain them.
        """
        self._errors = ""
        self._name = name
        self._other_args = other_args
        self._schema = schema
        self._schema_format = schema_format
        self._schema_conditionals = schema_conditionals
        self._data = data

    def _extract_schema_from_doc(self, doc_obj, temp_schema):
        """Extract the schema from a doc string
        :param doc_obj: The doc as a python obj
        :type doc_obj: dictionary
        :params temp_schema: The dict in which we stuff the schema parts
        :type temp_schema: dict
        """
        options_obj = doc_obj.get("options")
        for okey, ovalue in options_obj.items():
            temp_schema[okey] = {}
            for metakey in list(ovalue):
                if metakey == "suboptions":
                    temp_schema[okey].update({"options": {}})
                    suboptions_obj = {"options": ovalue["suboptions"]}
                    self._extract_schema_from_doc(suboptions_obj, temp_schema[okey]["options"])
                elif metakey in OPTION_METADATA + OPTION_CONDITIONALS:
                    temp_schema[okey].update({metakey: ovalue[metakey]})

    # TODO: Support extends_documentation_fragment
    def _convert_doc_to_schema(self):
        """Convert the doc string to an obj, was yaml
        add back other valid conditionals and params
        """
        doc_obj = yaml.load(self._schema, SafeLoader)
        temp_schema = {}
        self._extract_schema_from_doc(doc_obj, temp_schema)
        self._schema = {"argument_spec": temp_schema}

    def _validate(self):
        """Validate the data gainst the schema
        convert doc string in argspec if necessary

        :return valid: if the data passed
        :rtype valid: bool
        :return errors: errors reported during validation
        :rtype errors: str
        :return params: The original data updated with defaults
        :rtype params: dict
        """
        if self._schema_format == "doc":
            self._convert_doc_to_schema()
        if self._schema_conditionals is not None:
            self._schema = dict_merge(self._schema, self._schema_conditionals)
        if self._other_args is not None:
            self._schema = dict_merge(self._schema, self._other_args)
        invalid_keys = [k for k in self._schema.keys() if k not in VALID_ANSIBLEMODULE_ARGS]
        if invalid_keys:
            valid = False
            errors = "Invalid schema. Invalid keys found: {ikeys}".format(
                ikeys=",".join(invalid_keys),
            )
            updated_data = {}
        else:
            mm = MonkeyModule(data=self._data, schema=self._schema, name=self._name)
            valid, errors, updated_data = mm.validate()
        return valid, errors, updated_data

    def validate(self):
        """The public validate method
        check for future argspec validation
        that is coming in 2.11, change the check according above
        """
        if HAS_ANSIBLE_ARG_SPEC_VALIDATOR:
            if self._schema_format == "doc":
                self._convert_doc_to_schema()
            if self._schema_conditionals is not None:
                self._schema = dict_merge(self._schema, self._schema_conditionals)
            invalid_keys = [k for k in self._schema.keys() if k not in VALID_ANSIBLEMODULE_ARGS]
            if invalid_keys:
                valid = False
                errors = [
                    "Invalid schema. Invalid keys found: {ikeys}".format(
                        ikeys=",".join(invalid_keys),
                    ),
                ]
                updated_data = {}
                return valid, errors, updated_data
            else:
                validator = ArgumentSpecValidator(**self._schema)
                result = validator.validate(self._data)
                valid = not bool(result.error_messages)
                return (
                    valid,
                    result.error_messages,
                    result.validated_parameters,
                )
        else:
            return self._validate()


def check_argspec(schema, name, schema_format="doc", schema_conditionals=None, **args):
    if schema_conditionals is None:
        schema_conditionals = {}

    aav = AnsibleArgSpecValidator(
        data=args,
        schema=schema,
        schema_format=schema_format,
        schema_conditionals=schema_conditionals,
        name=name,
    )
    result = {}
    valid, errors, updated_params = aav.validate()
    if not valid:
        result["errors"] = errors
        result["failed"] = True
        result["msg"] = "argspec validation failed for {name} plugin".format(name=name)

    return valid, result, updated_params
